((nil
  (bug-reference-bug-regexp . "\\(\\(?:issue\\|pr\\) ?\\)?#\\([[:digit:]]+\\)")
  (bug-reference-url-format . "https://github.com/magnars/dash.el/issues/%s")
  (fill-column . 70)
  (sentence-end-double-space . t)
  (tab-width . 8))
 (emacs-lisp-mode
  (indent-tabs-mode . nil)
  (mode . bug-reference-prog))
 (sh-mode
  (sh-basic-offset . 4))
 (texinfo-mode
  (indent-tabs-mode . nil)
  (mode . bug-reference-prog)))
