;; Generated package description from dash.el  -*- no-byte-compile: t -*-
(define-package "dash" "2.19.0.0.20210708.200927" "A modern list library for Emacs" '((emacs "24")) :keywords '("extensions" "lisp") :authors '(("Magnar Sveen" . "magnars@gmail.com")) :maintainer '("Magnar Sveen" . "magnars@gmail.com") :url "https://github.com/magnars/dash.el")
